// qth-locator  Copyright 2019 Juuso Lehtinen https://github.com/jleh/qth-locator/  MIT License
'use strict';

var CHAR_CODE_OFFSET = 65;

var isValidLocatorString = function isValidLocatorString(locatorString) {
  return locatorString.match(/^[A-Ra-r][A-Ra-r]\d\d[A-Xa-x][A-Xa-x]/) !== null;
};

var charToNumber = function charToNumber(_char) {
  return _char.toUpperCase().charCodeAt(0) - CHAR_CODE_OFFSET;
};

var numberToChar = function numberToChar(number) {
  return String.fromCharCode(number + CHAR_CODE_OFFSET);
};

var locatorToLatLng = function locatorToLatLng(locatorString) {
  locatorString += 'll'; // append subsquare in case is 4 chars long...  If not, is ignored.

  if (!isValidLocatorString(locatorString)) {
    throw new Error('Input is not valid locator string');
  }

  var squareLng = charToNumber(locatorString[0]) * 20;
  var squareLat = charToNumber(locatorString[1]) * 10;
  var gridLng = Number.parseInt(locatorString[2]) * 2;
  var gridLat = Number.parseInt(locatorString[3]);
  var subsquareLng = (charToNumber(locatorString[4]) + 0.5) / 12;
  var subsquareLat = (charToNumber(locatorString[5]) + 0.5) / 24;
  return [squareLat + gridLat + subsquareLat - 90, squareLng + gridLng + subsquareLng - 180];
};

var degToRad = function degToRad(deg) {
  return deg % 360 * Math.PI / 180;
};

var radToDeg = function radToDeg(rad) {
  return rad / Math.PI * 180 % 360;
};

var bearingDistance = function bearingDistance(from, to) {
  var fromCoords = locatorToLatLng(from);
  var toCoords = locatorToLatLng(to);
  var dLat = degToRad(toCoords[0] - fromCoords[0]);
  var dLon = degToRad(toCoords[1] - fromCoords[1]);
  var fromLat = degToRad(fromCoords[0]);
  var toLat = degToRad(toCoords[0]);
  var a = Math.pow(Math.sin(dLat / 2), 2) + Math.pow(Math.sin(dLon / 2), 2) * Math.cos(fromLat) * Math.cos(toLat);
  var b = 2 * Math.atan2(Math.sqrt(a), Math.sqrt(1 - a));
  var y = dLon * Math.cos(fromLat) * Math.cos(toLat);
  var x = Math.sin(toLat) - Math.sin(fromLat) * Math.cos(b);
  var az = Math.atan2(y, x);

  if (az < 0) {
    az += 2 * Math.PI;
  }

  return {
    km: b * 6371,
    deg: radToDeg(az)
  };
};

var distance = function distance(from, to) {
  return bearingDistance(from, to).km;
};

var isValidPoint = function isValidPoint(lat, lng) {
  return lat >= -90 && lat <= 90 && lng >= -180 && lng <= 180;
};

var latLngToLocator = function latLngToLocator(lat, lng) {
  if (!isValidPoint(lat, lng)) {
    throw new Error('Input is not a valid coordinate');
  }

  var longitude = lng + 180;
  var latitude = lat + 90;
  var squareLng = numberToChar(Math.floor(longitude / 20));
  var squareLat = numberToChar(Math.floor(latitude / 10));
  var fieldLng = Math.floor(longitude % 20 / 2);
  var fieldLat = Math.floor(latitude % 10);
  var subsquareLng = numberToChar(Math.floor(longitude % 20 % 2 * 12)).toLowerCase();
  var subsquareLat = numberToChar((latitude % 10 - fieldLat) * 24).toLowerCase();
  return squareLng + squareLat + fieldLng + fieldLat + subsquareLng + subsquareLat;
};

module.exports = {
  isValidLocatorString: isValidLocatorString,
  locatorToLatLng: locatorToLatLng,
  distance: distance,
  bearingDistance: bearingDistance,
  latLngToLocator: latLngToLocator
};